"""
django-cotton

App configuration to set up the cotton loader and builtins automatically.
"""
import re
from contextlib import suppress

import django.contrib.admin
import django.template
from django.apps import AppConfig
from django.conf import settings


def wrap_loaders(name):
    for template_config in settings.TEMPLATES:
        engine_name = template_config.get("NAME")
        if not engine_name:
            engine_name = template_config["BACKEND"].split(".")[-2]
        if engine_name == name:
            loaders = template_config.setdefault("OPTIONS", {}).get("loaders", [])

            loaders_already_configured = (
                loaders
                and isinstance(loaders, (list, tuple))
                and isinstance(loaders[0], (tuple, list))
                and loaders[0][0] == "django.template.loaders.cached.Loader"
                and "django_cotton.cotton_loader.Loader" in loaders[0][1]
            )

            if not loaders_already_configured:
                template_config.pop("APP_DIRS", None)
                default_loaders = [
                    "django_cotton.cotton_loader.Loader",
                    "django.template.loaders.filesystem.Loader",
                    "django.template.loaders.app_directories.Loader",
                ]
                cached_loaders = [("django.template.loaders.cached.Loader", default_loaders)]
                template_config["OPTIONS"]["loaders"] = cached_loaders

            options = template_config.setdefault("OPTIONS", {})
            builtins = options.setdefault("builtins", [])
            builtins_already_configured = (
                builtins and "django_cotton.templatetags.cotton" in builtins
            )
            if not builtins_already_configured:
                template_config["OPTIONS"]["builtins"].insert(
                    0, "django_cotton.templatetags.cotton"
                )

            break

    # Force re-evaluation of settings.TEMPLATES because EngineHandler caches it.
    with suppress(AttributeError):
        del django.template.engines.templates
        django.template.engines._engines = {}


class LoaderAppConfig(AppConfig):
    """
    This, the default configuration, does the automatic setup of a partials loader.
    """

    name = "django_cotton"
    default = True

    def ready(self):
        from django.template import base

        # Support for multiline tags
        base.tag_re = re.compile(base.tag_re.pattern, re.DOTALL)

        wrap_loaders("django")


class SimpleAppConfig(AppConfig):
    """
    This, the non-default configuration, allows the user to opt-out of the automatic configuration. They just need to
    add "django_cotton.apps.SimpleAppConfig" to INSTALLED_APPS instead of "django_cotton".
    """

    name = "django_cotton"

    def ready(self):
        from django.template import base

        # Support for multiline tags
        base.tag_re = re.compile(base.tag_re.pattern, re.DOTALL)



import functools
from typing import Union

from django.conf import settings
from django.template import Library
from django.template.base import (
    Node,
)
from django.template.context import Context
from django.template.loader import get_template

from django_cotton.utils import get_cotton_data
from django_cotton.exceptions import CottonIncompleteDynamicComponentError
from django_cotton.templatetags import Attrs, DynamicAttr, UnprocessableDynamicAttr

register = Library()


class CottonComponentNode(Node):
    def __init__(self, component_name, nodelist, attrs, only):
        self.component_name = component_name
        self.nodelist = nodelist
        self.attrs = attrs
        self.template_cache = {}
        self.only = only

    def render(self, context):
        cotton_data = get_cotton_data(context)

        # Push a new component onto the stack
        component_data = {
            "key": self.component_name,
            "attrs": Attrs({}),
            "slots": {},
        }
        cotton_data["stack"].append(component_data)

        # Process simple attributes and boolean attributes
        for key, value in self.attrs.items():
            value = self._strip_quotes_safely(value)
            if value is True:  # Boolean attribute
                component_data["attrs"][key] = True
            elif key.startswith("::"):  # Escaping 1 colon e.g for shorthand alpine
                key = key[1:]
                # component_data["slots"][key] = value
                component_data["attrs"][key] = value
            elif key.startswith(":"):
                key = key[1:]
                try:
                    component_data["attrs"][key] = DynamicAttr(value).resolve(context)
                except UnprocessableDynamicAttr:
                    component_data["attrs"].unprocessable(key)
            else:
                component_data["attrs"][key] = value

        # Render the nodelist to process any slot tags and vars
        default_slot = self.nodelist.render(context)

        # Prepare the cotton-specific data
        component_state = {
            **component_data["slots"],
            **component_data["attrs"].make_attrs_accessible(),
            "attrs": component_data["attrs"],
            "slot": default_slot,
            "cotton_data": cotton_data,
        }

        template = self._get_cached_template(context, component_data["attrs"])

        # Isolate context if needed
        if self.only:
            output = template.render(Context(component_state))
        else:
            with context.push(component_state):
                output = template.render(context)

        cotton_data["stack"].pop()

        return output

    def _get_cached_template(self, context, attrs):
        cache = context.render_context.get(self)
        if cache is None:
            cache = context.render_context[self] = {}

        template_path = self._generate_component_template_path(self.component_name, attrs.get("is"))

        if template_path not in cache:
            template = get_template(template_path)
            if hasattr(template, "template"):
                template = template.template
            cache[template_path] = template

        return cache[template_path]

    @staticmethod
    @functools.lru_cache(maxsize=400)
    def _generate_component_template_path(component_name: str, is_: Union[str, None]) -> str:
        """Generate the path to the template for the given component name."""
        if component_name == "component":
            if is_ is None:
                raise CottonIncompleteDynamicComponentError(
                    'Cotton error: "<c-component>" should be accompanied by an "is" attribute.'
                )
            component_name = is_

        component_tpl_path = component_name.replace(".", "/")

        # Cotton by default will look for snake_case version of comp names. This can be configured to allow hyphenated names.
        snaked_cased_named = getattr(settings, "COTTON_SNAKE_CASED_NAMES", True)
        if snaked_cased_named:
            component_tpl_path = component_tpl_path.replace("-", "_")

        cotton_dir = getattr(settings, "COTTON_DIR", "cotton")
        return f"{cotton_dir}/{component_tpl_path}.html"

    @staticmethod
    def _strip_quotes_safely(value):
        if type(value) is str and value.startswith('"') and value.endswith('"'):
            return value[1:-1]
        return value


def cotton_component(parser, token):
    bits = token.split_contents()[1:]
    component_name = bits[0]
    attrs = {}
    only = False

    node_class = CottonComponentNode

    for bit in bits[1:]:
        if bit == "only":
            # if we see `only` we isolate context
            only = True
            continue
        try:
            key, value = bit.split("=")
            attrs[key] = value
        except ValueError:
            attrs[bit] = True

    nodelist = parser.parse(("endc",))
    parser.delete_first_token()

    return node_class(component_name, nodelist, attrs, only)



from django.template import Library
from django.template.base import (
    Node,
    TemplateSyntaxError,
)
from django.utils.safestring import mark_safe

from django_cotton.utils import get_cotton_data

register = Library()


def cotton_slot(parser, token):
    bits = token.split_contents()[1:]
    if len(bits) < 1:
        raise TemplateSyntaxError("cotton slot tag must include a 'name'")

    nodelist = parser.parse(("endslot",))
    parser.delete_first_token()
    return CottonSlotNode(bits[0], nodelist)


class CottonSlotNode(Node):
    def __init__(self, slot_name, nodelist):
        self.slot_name = slot_name
        self.nodelist = nodelist

    def render(self, context):
        cotton_data = get_cotton_data(context)
        if cotton_data["stack"]:
            content = self.nodelist.render(context)
            cotton_data["stack"][-1]["slots"][self.slot_name] = mark_safe(content)
        else:
            raise TemplateSyntaxError("<slot /> tag must be used inside a component")
        return ""



import ast
from collections.abc import Mapping
from typing import Set, Any, Dict, List

from django.template import Variable, TemplateSyntaxError, Context
from django.template.base import VariableDoesNotExist, Template
from django.utils.safestring import mark_safe

from django_cotton.utils import ensure_quoted


class UnprocessableDynamicAttr(Exception):
    pass


class DynamicAttr:
    def __init__(self, value: str, is_cvar=False):
        self.value = value
        self._is_cvar = is_cvar
        self._resolved_value = None

    def resolve(self, context: Context) -> Any:
        if self._resolved_value is not None:
            return self._resolved_value

        resolvers = [
            self._resolve_as_variable,
            self._resolve_as_boolean,
            self._resolve_as_template,
            self._resolve_as_literal,
        ]

        for resolver in resolvers:
            try:
                # noinspection PyArgumentList
                self._resolved_value = resolver(context)
                return self._resolved_value
            except (VariableDoesNotExist, TemplateSyntaxError, ValueError, SyntaxError):
                continue

        raise UnprocessableDynamicAttr

    def _resolve_as_variable(self, context):
        return Variable(self.value).resolve(context)

    def _resolve_as_boolean(self, _):
        if self.value == "":
            return True
        raise ValueError

    def _resolve_as_template(self, context):
        template = Template(self.value)
        rendered_value = template.render(context)
        if rendered_value != self.value:
            return rendered_value
        raise TemplateSyntaxError

    def _resolve_as_literal(self, _):
        return ast.literal_eval(self.value)


class Attrs(Mapping):
    def __init__(self, attrs: Dict[str, Any]):
        self._attrs = attrs
        self._exclude_from_str: Set[str] = set()
        self._unprocessable: List[str] = []

    def __str__(self):
        return mark_safe(
            " ".join(
                f"{k}" if v is True else f"{k}={ensure_quoted(v)}"
                for k, v in self._attrs.items()
                if k not in self._exclude_from_str
            )
        )

    def __getitem__(self, key):
        return self._attrs[key]

    def __setitem__(self, key, value):
        self._attrs[key] = value

    def __iter__(self):
        return iter(self._attrs)

    def __len__(self):
        return len(self._attrs)

    def items(self):
        return self._attrs.items()

    def keys(self):
        return self._attrs.keys()

    def values(self):
        return self._attrs.values()

    # Custom methods to allow modifications
    @property
    def dict(self):
        return self._attrs

    def unprocessable(self, key):
        self._unprocessable.append(key)

    def exclude_unprocessable(self):
        return {k: v for k, v in self._attrs.items() if k not in self._unprocessable}

    def exclude_from_string_output(self, key):
        self._exclude_from_str.add(key)

    def make_attrs_accessible(self):
        return {k.replace("-", "_"): v for k, v in self._attrs.items()}



from django.shortcuts import render


# benchmarks
def compiled_cotton_test_view(request):
    return render(request, "compiled_cotton_test.html")


def native_extends_test_view(request):
    return render(request, "native_extends_test.html")


def native_include_test_view(request):
    return render(request, "native_include_test.html")
